@@ -19,9 +19,9 @@
 
 #include <apt-pkg/configuration.h>
 #include <apt-pkg/error.h>
-#include <apt-pkg/strutl.h>
 #include <apt-pkg/fileutl.h>
 #include <apt-pkg/macros.h>
+#include <apt-pkg/strutl.h>
 
 #include <ctype.h>
 #include <regex.h>
@@ -29,10 +29,15 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+
 #include <algorithm>
-#include <string>
-#include <vector>
 #include <fstream>
+#include <iterator>
+#include <sstream>
+#include <stack>
+#include <string>
+#include <unordered_map>
+#include <vector>
 
 #include <apti18n.h>
 
@@ -41,6 +46,155 @@ using namespace std;
 
 Configuration *_config = new Configuration;
 
+/* TODO: This config verification shouldn't be using a static variable
+   but a Cnf-member – but that would need ABI breaks and stuff and for now
+   that really is an apt-dev-only tool, so it isn't that bad that it is
+   unusable and allaround a bit strange */
+enum class APT_HIDDEN ConfigType { UNDEFINED, INT, BOOL, STRING, STRING_OR_BOOL, STRING_OR_LIST, FILE, DIR, LIST, PROGRAM_PATH = FILE };
+APT_HIDDEN std::unordered_map<std::string, ConfigType> apt_known_config {};
+static std::string getConfigTypeString(ConfigType const type)		/*{{{*/
+{
+   switch (type)
+   {
+      case ConfigType::UNDEFINED: return "UNDEFINED";
+      case ConfigType::INT: return "INT";
+      case ConfigType::BOOL: return "BOOL";
+      case ConfigType::STRING: return "STRING";
+      case ConfigType::STRING_OR_BOOL: return "STRING_OR_BOOL";
+      case ConfigType::FILE: return "FILE";
+      case ConfigType::DIR: return "DIR";
+      case ConfigType::LIST: return "LIST";
+      case ConfigType::STRING_OR_LIST: return "STRING_OR_LIST";
+   }
+   return "UNKNOWN";
+}
+									/*}}}*/
+static ConfigType getConfigType(std::string const &type)		/*{{{*/
+{
+   if (type == "<INT>")
+      return ConfigType::INT;
+   else if (type == "<BOOL>")
+      return ConfigType::BOOL;
+   else if (type == "<STRING>")
+      return ConfigType::STRING;
+   else if (type == "<STRING_OR_BOOL>")
+      return ConfigType::STRING_OR_BOOL;
+   else if (type == "<FILE>")
+      return ConfigType::FILE;
+   else if (type == "<DIR>")
+      return ConfigType::DIR;
+   else if (type == "<LIST>")
+      return ConfigType::LIST;
+   else if (type == "<STRING_OR_LIST>")
+      return ConfigType::STRING_OR_LIST;
+   else if (type == "<PROGRAM_PATH>")
+      return ConfigType::PROGRAM_PATH;
+   return ConfigType::UNDEFINED;
+}
+									/*}}}*/
+// checkFindConfigOptionType - workhorse of option checking		/*{{{*/
+static void checkFindConfigOptionTypeInternal(std::string name, ConfigType const type)
+{
+   std::transform(name.begin(), name.end(), name.begin(), ::tolower);
+   auto known = apt_known_config.find(name);
+   if (known == apt_known_config.cend())
+   {
+      auto const rcolon = name.rfind(':');
+      if (rcolon != std::string::npos)
+      {
+	 known = apt_known_config.find(name.substr(0, rcolon) + ":*");
+	 if (known == apt_known_config.cend())
+	 {
+	    auto const parts = StringSplit(name, "::");
+	    size_t psize = parts.size();
+	    if (psize > 1)
+	    {
+	       for (size_t max = psize; max != 1; --max)
+	       {
+		  std::ostringstream os;
+		  std::copy(parts.begin(), parts.begin() + max, std::ostream_iterator<std::string>(os, "::"));
+		  os << "**";
+		  known = apt_known_config.find(os.str());
+		  if (known != apt_known_config.cend() && known->second == ConfigType::UNDEFINED)
+		     return;
+	       }
+	       for (size_t max = psize - 1; max != 1; --max)
+	       {
+		  std::ostringstream os;
+		  std::copy(parts.begin(), parts.begin() + max - 1, std::ostream_iterator<std::string>(os, "::"));
+		  os << "*::";
+		  std::copy(parts.begin() + max + 1, parts.end() - 1, std::ostream_iterator<std::string>(os, "::"));
+		  os << *(parts.end() - 1);
+		  known = apt_known_config.find(os.str());
+		  if (known != apt_known_config.cend())
+		     break;
+	       }
+	    }
+	 }
+      }
+   }
+   if (known == apt_known_config.cend())
+      _error->Warning("Using unknown config option »%s« of type %s",
+	    name.c_str(), getConfigTypeString(type).c_str());
+   else if (known->second != type)
+   {
+      if (known->second == ConfigType::DIR && type == ConfigType::FILE)
+	 ; // implementation detail
+      else if (type == ConfigType::STRING && (known->second == ConfigType::FILE || known->second == ConfigType::DIR))
+	 ; // TODO: that might be an error or not, we will figure this out later
+      else if (known->second == ConfigType::STRING_OR_BOOL && (type == ConfigType::BOOL || type == ConfigType::STRING))
+	 ;
+      else if (known->second == ConfigType::STRING_OR_LIST && (type == ConfigType::LIST || type == ConfigType::STRING))
+	 ;
+      else
+	 _error->Warning("Using config option »%s« of type %s as a type %s",
+	       name.c_str(), getConfigTypeString(known->second).c_str(), getConfigTypeString(type).c_str());
+   }
+}
+static void checkFindConfigOptionType(char const * const name, ConfigType const type)
+{
+   if (apt_known_config.empty())
+      return;
+   checkFindConfigOptionTypeInternal(name, type);
+}
+									/*}}}*/
+static bool LoadConfigurationIndex(std::string const &filename)		/*{{{*/
+{
+   apt_known_config.clear();
+   if (filename.empty())
+      return true;
+   Configuration Idx;
+   if (ReadConfigFile(Idx, filename) == false)
+      return false;
+
+   Configuration::Item const * Top = Idx.Tree(nullptr);
+   if (unlikely(Top == nullptr))
+      return false;
+
+   do {
+      if (Top->Value.empty() == false)
+      {
+	 std::string fulltag = Top->FullTag();
+	 std::transform(fulltag.begin(), fulltag.end(), fulltag.begin(), ::tolower);
+	 apt_known_config.emplace(std::move(fulltag), getConfigType(Top->Value));
+      }
+
+      if (Top->Child != nullptr)
+      {
+	 Top = Top->Child;
+	 continue;
+      }
+
+      while (Top != nullptr && Top->Next == nullptr)
+	 Top = Top->Parent;
+      if (Top != nullptr)
+	 Top = Top->Next;
+   } while (Top != nullptr);
+
+   return true;
+}
+									/*}}}*/
+
 // Configuration::Configuration - Constructor				/*{{{*/
 // ---------------------------------------------------------------------
 /* */
@@ -158,6 +312,7 @@ Configuration::Item *Configuration::Lookup(const char *Name,bool const &Create)
 /* */
 string Configuration::Find(const char *Name,const char *Default) const
 {
+   checkFindConfigOptionType(Name, ConfigType::STRING);
    const Item *Itm = Lookup(Name);
    if (Itm == 0 || Itm->Value.empty() == true)
    {
@@ -177,6 +332,7 @@ string Configuration::Find(const char *Name,const char *Default) const
  */
 string Configuration::FindFile(const char *Name,const char *Default) const
 {
+   checkFindConfigOptionType(Name, ConfigType::FILE);
    const Item *RootItem = Lookup("RootDir");
    std::string result =  (RootItem == 0) ? "" : RootItem->Value;
    if(result.empty() == false && result[result.size() - 1] != '/')
@@ -223,15 +379,7 @@ string Configuration::FindFile(const char *Name,const char *Default) const
       }
       result.append(val);
    }
-
-   // do some normalisation by removing // and /./ from the path
-   size_t found = string::npos;
-   while ((found = result.find("/./")) != string::npos)
-      result.replace(found, 3, "/");
-   while ((found = result.find("//")) != string::npos)
-      result.replace(found, 2, "/");
-
-   return result;
+   return flNormalize(result);
 }
 									/*}}}*/
 // Configuration::FindDir - Find a directory name			/*{{{*/
@@ -239,6 +387,7 @@ string Configuration::FindFile(const char *Name,const char *Default) const
 /* This is like findfile execept the result is terminated in a / */
 string Configuration::FindDir(const char *Name,const char *Default) const
 {
+   checkFindConfigOptionType(Name, ConfigType::DIR);
    string Res = FindFile(Name,Default);
    if (Res.end()[-1] != '/')
    {
@@ -253,11 +402,9 @@ string Configuration::FindDir(const char *Name,const char *Default) const
 // Configuration::FindVector - Find a vector of values			/*{{{*/
 // ---------------------------------------------------------------------
 /* Returns a vector of config values under the given item */
-#if (APT_PKG_MAJOR >= 4 && APT_PKG_MINOR < 13)
-vector<string> Configuration::FindVector(const char *Name) const { return FindVector(Name, ""); }
-#endif
-vector<string> Configuration::FindVector(const char *Name, std::string const &Default) const
+vector<string> Configuration::FindVector(const char *Name, std::string const &Default, bool const Keys) const
 {
+   checkFindConfigOptionType(Name, ConfigType::LIST);
    vector<string> Vec;
    const Item *Top = Lookup(Name);
    if (Top == NULL)
@@ -269,7 +416,7 @@ vector<string> Configuration::FindVector(const char *Name, std::string const &De
    Item *I = Top->Child;
    while(I != NULL)
    {
-      Vec.push_back(I->Value);
+      Vec.push_back(Keys ? I->Tag : I->Value);
       I = I->Next;
    }
    if (Vec.empty() == true)
@@ -283,6 +430,7 @@ vector<string> Configuration::FindVector(const char *Name, std::string const &De
 /* */
 int Configuration::FindI(const char *Name,int const &Default) const
 {
+   checkFindConfigOptionType(Name, ConfigType::INT);
    const Item *Itm = Lookup(Name);
    if (Itm == 0 || Itm->Value.empty() == true)
       return Default;
@@ -300,6 +448,7 @@ int Configuration::FindI(const char *Name,int const &Default) const
 /* */
 bool Configuration::FindB(const char *Name,bool const &Default) const
 {
+   checkFindConfigOptionType(Name, ConfigType::BOOL);
    const Item *Itm = Lookup(Name);
    if (Itm == 0 || Itm->Value.empty() == true)
       return Default;
@@ -488,6 +637,59 @@ void Configuration::Clear(string const &Name)
    }
 }
 									/*}}}*/
+void Configuration::MoveSubTree(char const * const OldRootName, char const * const NewRootName)/*{{{*/
+{
+   // prevent NewRoot being a subtree of OldRoot
+   if (OldRootName == nullptr)
+      return;
+   if (NewRootName != nullptr)
+   {
+      if (strcmp(OldRootName, NewRootName) == 0)
+	 return;
+      std::string const oldroot = std::string(OldRootName) + "::";
+      if (strcasestr(NewRootName, oldroot.c_str()) != NULL)
+	 return;
+   }
+
+   Item * Top;
+   Item const * const OldRoot = Top = Lookup(OldRootName, false);
+   if (Top == nullptr)
+      return;
+   std::string NewRoot;
+   if (NewRootName != nullptr)
+      NewRoot.append(NewRootName).append("::");
+
+   Top->Value.clear();
+   Item * const Stop = Top;
+   Top = Top->Child;
+   Stop->Child = 0;
+   for (; Top != 0;)
+   {
+      if (Top->Child != 0)
+      {
+	 Top = Top->Child;
+	 continue;
+      }
+
+      while (Top != 0 && Top->Next == 0)
+      {
+	 Set(NewRoot + Top->FullTag(OldRoot), Top->Value);
+	 Item const * const Tmp = Top;
+	 Top = Top->Parent;
+	 delete Tmp;
+
+	 if (Top == Stop)
+	    return;
+      }
+
+      Set(NewRoot + Top->FullTag(OldRoot), Top->Value);
+      Item const * const Tmp = Top;
+      if (Top != 0)
+	 Top = Top->Next;
+      delete Tmp;
+   }
+}
+									/*}}}*/
 // Configuration::Exists - Returns true if the Name exists		/*{{{*/
 // ---------------------------------------------------------------------
 /* */
@@ -624,43 +826,41 @@ string Configuration::Item::FullTag(const Item *Stop) const
    sections like 'zone "foo.org" { .. };' This causes each section to be
    added in with a tag like "zone::foo.org" instead of being split 
    tag/value. AsSectional enables Sectional parsing.*/
+static void leaveCurrentScope(std::stack<std::string> &Stack, std::string &ParentTag)
+{
+   if (Stack.empty())
+      ParentTag.clear();
+   else
+   {
+      ParentTag = Stack.top();
+      Stack.pop();
+   }
+}
 bool ReadConfigFile(Configuration &Conf,const string &FName,bool const &AsSectional,
 		    unsigned const &Depth)
-{   
+{
    // Open the stream for reading
-   ifstream F(FName.c_str(),ios::in); 
-   if (!F != 0)
-      return _error->Errno("ifstream::ifstream",_("Opening configuration file %s"),FName.c_str());
+   FileFd F;
+   if (OpenConfigurationFileFd(FName, F) == false)
+      return false;
 
    string LineBuffer;
-   string Stack[100];
-   unsigned int StackPos = 0;
-   
+   std::stack<std::string> Stack;
+
    // Parser state
    string ParentTag;
-   
+
    int CurLine = 0;
    bool InComment = false;
-   while (F.eof() == false)
+   while (F.Eof() == false)
    {
       // The raw input line.
       std::string Input;
+      if (F.ReadLine(Input) == false)
+	 Input.clear();
       // The input line with comments stripped.
       std::string Fragment;
 
-      // Grab the next line of F and place it in Input.
-      do
-	{
-	  char *Buffer = new char[1024];
-
-	  F.clear();
-	  F.getline(Buffer,sizeof(Buffer) / 2);
-
-	  Input += Buffer;
-	  delete[] Buffer;
-	}
-      while (F.fail() && !F.eof());
-
       // Expand tabs in the input line and remove leading and trailing
       // whitespace.
       {
@@ -703,12 +903,12 @@ bool ReadConfigFile(Configuration &Conf,const string &FName,bool const &AsSectio
 	       Start = I + 2;
 	       InComment = false;
 	       break;
-	    }	    
+	    }
 	 }
 	 if (InComment == true)
 	    continue;
       }
-      
+
       // Discard single line comments
       bool InQuote = false;
       for (std::string::const_iterator I = Start;
@@ -721,7 +921,8 @@ bool ReadConfigFile(Configuration &Conf,const string &FName,bool const &AsSectio
 
 	 if ((*I == '/' && I + 1 != End && I[1] == '/') ||
 	     (*I == '#' && strcmp(string(I,I+6).c_str(),"#clear") != 0 &&
-	      strcmp(string(I,I+8).c_str(),"#include") != 0))
+	      strcmp(string(I,I+8).c_str(),"#include") != 0 &&
+	      strcmp(string(I,I+strlen("#x-apt-configure-index")).c_str(), "#x-apt-configure-index") != 0))
 	 {
 	    End = I;
 	    break;
@@ -753,9 +954,9 @@ bool ReadConfigFile(Configuration &Conf,const string &FName,bool const &AsSectio
 		  I = J + 1;
 		  InComment = false;
 		  break;
-	       }	       
+	       }
 	    }
-	    
+
 	    if (InComment == true)
 	      break;
 	 }
@@ -766,7 +967,7 @@ bool ReadConfigFile(Configuration &Conf,const string &FName,bool const &AsSectio
       // Skip blank lines.
       if (Fragment.empty())
 	 continue;
-      
+
       // The line has actual content; interpret what it means.
       InQuote = false;
       Start = Fragment.begin();
@@ -776,7 +977,7 @@ bool ReadConfigFile(Configuration &Conf,const string &FName,bool const &AsSectio
       {
 	 if (*I == '"')
 	    InQuote = !InQuote;
-	 
+
 	 if (InQuote == false && (*I == '{' || *I == ';' || *I == '}'))
 	 {
 	    // Put the last fragment into the buffer
@@ -795,24 +996,19 @@ bool ReadConfigFile(Configuration &Conf,const string &FName,bool const &AsSectio
 	    // brace or a semicolon)
 	    char TermChar = *I;
 	    Start = I + 1;
-	    
+
 	    // Syntax Error
 	    if (TermChar == '{' && LineBuffer.empty() == true)
 	       return _error->Error(_("Syntax error %s:%u: Block starts with no name."),FName.c_str(),CurLine);
-	    
+
 	    // No string on this line
 	    if (LineBuffer.empty() == true)
 	    {
 	       if (TermChar == '}')
-	       {
-		  if (StackPos == 0)
-		     ParentTag = string();
-		  else
-		     ParentTag = Stack[--StackPos];
-	       }
+		  leaveCurrentScope(Stack, ParentTag);
 	       continue;
 	    }
-	    
+
 	    // Parse off the tag
 	    string Tag;
 	    const char *Pos = LineBuffer.c_str();
@@ -839,25 +1035,23 @@ bool ReadConfigFile(Configuration &Conf,const string &FName,bool const &AsSectio
 	    // Go down a level
 	    if (TermChar == '{')
 	    {
-	       if (StackPos < sizeof(Stack)/sizeof(std::string))
-		  Stack[StackPos++] = ParentTag;
-	       
-	       /* Make sectional tags incorperate the section into the
+	       Stack.push(ParentTag);
+
+	       /* Make sectional tags incorporate the section into the
 	          tag string */
 	       if (AsSectional == true && Word.empty() == false)
 	       {
-		  Tag += "::" ;
-		  Tag += Word;
-		  Word = "";
+		  Tag.append("::").append(Word);
+		  Word.clear();
 	       }
-	       
+
 	       if (ParentTag.empty() == true)
 		  ParentTag = Tag;
 	       else
-		  ParentTag += string("::") + Tag;
-	       Tag = string();
+		  ParentTag.append("::").append(Tag);
+	       Tag.clear();
 	    }
-	    
+
 	    // Generate the item name
 	    string Item;
 	    if (ParentTag.empty() == true)
@@ -869,7 +1063,7 @@ bool ReadConfigFile(Configuration &Conf,const string &FName,bool const &AsSectio
 	       else
 		  Item = ParentTag;
 	    }
-	    
+
 	    // Specials
 	    if (Tag.length() >= 1 && Tag[0] == '#')
 	    {
@@ -891,7 +1085,12 @@ bool ReadConfigFile(Configuration &Conf,const string &FName,bool const &AsSectio
 		  {
 		     if (ReadConfigFile(Conf,Word,AsSectional,Depth+1) == false)
 			return _error->Error(_("Syntax error %s:%u: Included from here"),FName.c_str(),CurLine);
-		  }		  
+		  }
+	       }
+	       else if (Tag == "x-apt-configure-index")
+	       {
+		  if (LoadConfigurationIndex(Word) == false)
+		     return _error->Warning("Loading the configure index %s in file %s:%u failed!", Word.c_str(), FName.c_str(), CurLine);
 	       }
 	       else
 		  return _error->Error(_("Syntax error %s:%u: Unsupported directive '%s'"),FName.c_str(),CurLine,Tag.c_str());
@@ -904,19 +1103,13 @@ bool ReadConfigFile(Configuration &Conf,const string &FName,bool const &AsSectio
 	       if (NoWord == false)
 		  Conf.Set(Item,Word);
 	    }
-	    
+
 	    // Empty the buffer
 	    LineBuffer.clear();
-	    
+
 	    // Move up a tag, but only if there is no bit to parse
 	    if (TermChar == '}')
-	    {
-	       if (StackPos == 0)
-		  ParentTag.clear();
-	       else
-		  ParentTag = Stack[--StackPos];
-	    }
-	    
+	       leaveCurrentScope(Stack, ParentTag);
 	 }
       }
 
@@ -957,13 +1150,10 @@ bool ReadConfigFile(Configuration &Conf,const string &FName,bool const &AsSectio
 bool ReadConfigDir(Configuration &Conf,const string &Dir,
 		   bool const &AsSectional, unsigned const &Depth)
 {
-   vector<string> const List = GetListOfFilesInDir(Dir, "conf", true, true);
-
-   // Read the files
-   for (vector<string>::const_iterator I = List.begin(); I != List.end(); ++I)
-      if (ReadConfigFile(Conf,*I,AsSectional,Depth) == false)
-	 return false;
-   return true;
+   bool good = true;
+   for (auto const &I : GetListOfFilesInDir(Dir, "conf", true, true))
+      good = ReadConfigFile(Conf, I, AsSectional, Depth) && good;
+   return good;
 }
 									/*}}}*/
 // MatchAgainstConfig Constructor					/*{{{*/
