@@ -1,24 +1,26 @@
 // -*- mode: cpp; mode: fold -*-
 // Include Files							/*{{{*/
-#include<config.h>
+#include <config.h>
 
-#include<apt-pkg/configuration.h>
-#include<apt-pkg/error.h>
-#include<apt-pkg/strutl.h>
-#include<apt-pkg/fileutl.h>
-#include<apt-pkg/gpgv.h>
+#include <apt-pkg/configuration.h>
+#include <apt-pkg/error.h>
+#include <apt-pkg/fileutl.h>
+#include <apt-pkg/gpgv.h>
+#include <apt-pkg/strutl.h>
 
 #include <errno.h>
-#include <stdio.h>
-#include <string.h>
-#include <stdlib.h>
 #include <fcntl.h>
+#include <stddef.h>
+#include <stdio.h>
+#include <stdlib.h>
+#include <string.h>
 #include <sys/wait.h>
 #include <unistd.h>
-#include <stddef.h>
 
 #include <algorithm>
+#include <fstream>
 #include <iostream>
+#include <sstream>
 #include <string>
 #include <vector>
 
@@ -43,13 +45,62 @@ static char * GenerateTemporaryFileTemplate(const char *basename)	/*{{{*/
    And as a cherry on the cake, we use our apt-key wrapper to do part
    of the lifting in regards to merging keyrings. Fun for the whole family.
 */
+static bool iovprintf(std::ostream &out, const char *format,
+		      va_list &args, ssize_t &size) {
+   char *S = (char*)malloc(size);
+   ssize_t const n = vsnprintf(S, size, format, args);
+   if (n > -1 && n < size) {
+      out << S;
+      free(S);
+      return true;
+   } else {
+      if (n > -1)
+	 size = n + 1;
+      else
+	 size *= 2;
+   }
+   free(S);
+   return false;
+}
+static void APT_PRINTF(4) apt_error(std::ostream &outterm, int const statusfd, int fd[2], const char *format, ...)
+{
+   std::ostringstream outstr;
+   std::ostream &out = (statusfd == -1) ? outterm : outstr;
+   va_list args;
+   ssize_t size = 400;
+   while (true) {
+      bool ret;
+      va_start(args,format);
+      ret = iovprintf(out, format, args, size);
+      va_end(args);
+      if (ret == true)
+	 break;
+   }
+   if (statusfd != -1)
+   {
+      auto const errtag = "[APTKEY:] ERROR ";
+      outstr << '\n';
+      auto const errtext = outstr.str();
+      if (FileFd::Write(fd[1], errtag, strlen(errtag)) == false ||
+	    FileFd::Write(fd[1], errtext.data(), errtext.size()) == false)
+	 outterm << errtext << std::flush;
+   }
+}
 void ExecGPGV(std::string const &File, std::string const &FileGPG,
              int const &statusfd, int fd[2], std::string const &key)
 {
    #define EINTERNAL 111
-   std::string const aptkey = _config->FindFile("Dir::Bin::apt-key", "/usr/bin/apt-key");
+   std::string const aptkey = _config->Find("Dir::Bin::apt-key", CMAKE_INSTALL_FULL_BINDIR "/apt-key");
 
    bool const Debug = _config->FindB("Debug::Acquire::gpgv", false);
+   struct exiter {
+      std::vector<const char *> files;
+      void operator ()(int code) APT_NORETURN {
+	 std::for_each(files.begin(), files.end(), unlink);
+	 exit(code);
+      }
+   } local_exit;
+
 
    std::vector<const char *> Args;
    Args.reserve(10);
@@ -94,9 +145,30 @@ void ExecGPGV(std::string const &File, std::string const &FileGPG,
    }
 
    enum  { DETACHED, CLEARSIGNED } releaseSignature = (FileGPG != File) ? DETACHED : CLEARSIGNED;
-   std::vector<std::string> dataHeader;
    char * sig = NULL;
    char * data = NULL;
+   char * conf = nullptr;
+
+   // Dump the configuration so apt-key picks up the correct Dir values
+   {
+      conf = GenerateTemporaryFileTemplate("apt.conf");
+      if (conf == nullptr) {
+	 apt_error(std::cerr, statusfd, fd, "Couldn't create tempfile names for passing config to apt-key");
+	 local_exit(EINTERNAL);
+      }
+      int confFd = mkstemp(conf);
+      if (confFd == -1) {
+	 apt_error(std::cerr, statusfd, fd, "Couldn't create temporary file %s for passing config to apt-key", conf);
+	 local_exit(EINTERNAL);
+      }
+      local_exit.files.push_back(conf);
+
+      std::ofstream confStream(conf);
+      close(confFd);
+      _config->Dump(confStream);
+      confStream.close();
+      setenv("APT_CONFIG", conf, 1);
+   }
 
    if (releaseSignature == DETACHED)
    {
@@ -109,20 +181,20 @@ void ExecGPGV(std::string const &File, std::string const &FileGPG,
       data = GenerateTemporaryFileTemplate("apt.data");
       if (sig == NULL || data == NULL)
       {
-	 ioprintf(std::cerr, "Couldn't create tempfile names for splitting up %s", File.c_str());
-	 exit(EINTERNAL);
+	 apt_error(std::cerr, statusfd, fd, "Couldn't create tempfile names for splitting up %s", File.c_str());
+	 local_exit(EINTERNAL);
       }
 
       int const sigFd = mkstemp(sig);
       int const dataFd = mkstemp(data);
+      if (dataFd != -1)
+	 local_exit.files.push_back(data);
+      if (sigFd != -1)
+	 local_exit.files.push_back(sig);
       if (sigFd == -1 || dataFd == -1)
       {
-	 if (dataFd != -1)
-	    unlink(data);
-	 if (sigFd != -1)
-	    unlink(sig);
-	 ioprintf(std::cerr, "Couldn't create tempfiles for splitting up %s", File.c_str());
-	 exit(EINTERNAL);
+	 apt_error(std::cerr, statusfd, fd, "Couldn't create tempfiles for splitting up %s", File.c_str());
+	 local_exit(EINTERNAL);
       }
 
       FileFd signature;
@@ -131,14 +203,10 @@ void ExecGPGV(std::string const &File, std::string const &FileGPG,
       message.OpenDescriptor(dataFd, FileFd::WriteOnly, true);
 
       if (signature.Failed() == true || message.Failed() == true ||
-	    SplitClearSignedFile(File, &message, &dataHeader, &signature) == false)
+	    SplitClearSignedFile(File, &message, nullptr, &signature) == false)
       {
-	 if (dataFd != -1)
-	    unlink(data);
-	 if (sigFd != -1)
-	    unlink(sig);
-	 ioprintf(std::cerr, "Splitting up %s into data and signature failed", File.c_str());
-	 exit(112);
+	 apt_error(std::cerr, statusfd, fd, "Splitting up %s into data and signature failed", File.c_str());
+	 local_exit(112);
       }
       Args.push_back(sig);
       Args.push_back(data);
@@ -171,67 +239,50 @@ void ExecGPGV(std::string const &File, std::string const &FileGPG,
       putenv((char *)"LC_MESSAGES=");
    }
 
-   if (releaseSignature == DETACHED)
+
+   // We have created tempfiles we have to clean up
+   // and we do an additional check, so fork yet another time …
+   pid_t pid = ExecFork();
+   if(pid < 0) {
+      apt_error(std::cerr, statusfd, fd, "Fork failed for %s to check %s", Args[0], File.c_str());
+      local_exit(EINTERNAL);
+   }
+   if(pid == 0)
    {
+      if (statusfd != -1)
+	 dup2(fd[1], statusfd);
       execvp(Args[0], (char **) &Args[0]);
-      ioprintf(std::cerr, "Couldn't execute %s to check %s", Args[0], File.c_str());
-      exit(EINTERNAL);
+      apt_error(std::cerr, statusfd, fd, "Couldn't execute %s to check %s", Args[0], File.c_str());
+      local_exit(EINTERNAL);
    }
-   else
+
+   // Wait and collect the error code - taken from WaitPid as we need the exact Status
+   int Status;
+   while (waitpid(pid,&Status,0) != pid)
    {
-//#define UNLINK_EXIT(X) exit(X)
-#define UNLINK_EXIT(X) unlink(sig);unlink(data);exit(X)
-
-      // for clear-signed files we have created tempfiles we have to clean up
-      // and we do an additional check, so fork yet another time …
-      pid_t pid = ExecFork();
-      if(pid < 0) {
-	 ioprintf(std::cerr, "Fork failed for %s to check %s", Args[0], File.c_str());
-	 UNLINK_EXIT(EINTERNAL);
-      }
-      if(pid == 0)
-      {
-	 if (statusfd != -1)
-	    dup2(fd[1], statusfd);
-	 execvp(Args[0], (char **) &Args[0]);
-	 ioprintf(std::cerr, "Couldn't execute %s to check %s", Args[0], File.c_str());
-	 UNLINK_EXIT(EINTERNAL);
-      }
-
-      // Wait and collect the error code - taken from WaitPid as we need the exact Status
-      int Status;
-      while (waitpid(pid,&Status,0) != pid)
-      {
-	 if (errno == EINTR)
-	    continue;
-	 ioprintf(std::cerr, _("Waited for %s but it wasn't there"), "apt-key");
-	 UNLINK_EXIT(EINTERNAL);
-      }
-#undef UNLINK_EXIT
-      // we don't need the files any longer
-      unlink(sig);
-      unlink(data);
-      free(sig);
-      free(data);
-
-      // check if it exit'ed normally …
-      if (WIFEXITED(Status) == false)
-      {
-	 ioprintf(std::cerr, _("Sub-process %s exited unexpectedly"), "apt-key");
-	 exit(EINTERNAL);
-      }
-
-      // … and with a good exit code
-      if (WEXITSTATUS(Status) != 0)
-      {
-	 ioprintf(std::cerr, _("Sub-process %s returned an error code (%u)"), "apt-key", WEXITSTATUS(Status));
-	 exit(WEXITSTATUS(Status));
-      }
-
-      // everything fine
-      exit(0);
+      if (errno == EINTR)
+	 continue;
+      apt_error(std::cerr, statusfd, fd, _("Waited for %s but it wasn't there"), "apt-key");
+      local_exit(EINTERNAL);
    }
-   exit(EINTERNAL); // unreachable safe-guard
+
+   // check if it exit'ed normally …
+   if (WIFEXITED(Status) == false)
+   {
+      apt_error(std::cerr, statusfd, fd, _("Sub-process %s exited unexpectedly"), "apt-key");
+      local_exit(EINTERNAL);
+   }
+
+   // … and with a good exit code
+   if (WEXITSTATUS(Status) != 0)
+   {
+      // we forward the statuscode, so don't generate a message on the fd in this case
+      apt_error(std::cerr, -1, fd, _("Sub-process %s returned an error code (%u)"), "apt-key", WEXITSTATUS(Status));
+      local_exit(WEXITSTATUS(Status));
+   }
+
+   // everything fine
+   local_exit(0);
 }
 									/*}}}*/
 // SplitClearSignedFile - split message into data/signature		/*{{{*/
@@ -261,6 +312,8 @@ bool SplitClearSignedFile(std::string const &InFile, FileFd * const ContentFile,
    bool skip_until_empty_line = false;
    bool found_signature = false;
    bool first_line = true;
+   bool signed_message_not_on_first_line = false;
+   bool found_garbage = false;
 
    char *buf = NULL;
    size_t buf_size = 0;
@@ -275,6 +328,8 @@ bool SplitClearSignedFile(std::string const &InFile, FileFd * const ContentFile,
 	    found_message_start = true;
 	    skip_until_empty_line = true;
 	 }
+	 else
+	    signed_message_not_on_first_line = found_garbage = true;
       }
       else if (skip_until_empty_line == true)
       {
@@ -312,6 +367,8 @@ bool SplitClearSignedFile(std::string const &InFile, FileFd * const ContentFile,
 	    if (ContentFile != NULL)
 	       ContentFile->Write(dashfree, strlen(dashfree));
 	 }
+	 else
+	    found_garbage = true;
       }
       else if (found_signature == true)
       {
@@ -324,6 +381,8 @@ bool SplitClearSignedFile(std::string const &InFile, FileFd * const ContentFile,
 	    found_signature = false; // look for other signatures
       }
       // all the rest is whitespace, unsigned garbage or additional message blocks we ignore
+      else
+	 found_garbage = true;
    }
    fclose(in);
    if (buf != NULL)
@@ -335,7 +394,15 @@ bool SplitClearSignedFile(std::string const &InFile, FileFd * const ContentFile,
    if (ContentFile != nullptr)
       ContentFile->Flush();
 
-   // An error occured during reading - propagate it up
+   if (found_message_start)
+   {
+      if (signed_message_not_on_first_line)
+	 _error->Warning("Clearsigned file '%s' does not start with a signed message block.", InFile.c_str());
+      else if (found_garbage)
+	 _error->Warning("Clearsigned file '%s' contains unsigned lines.", InFile.c_str());
+   }
+
+   // An error occurred during reading - propagate it up
    bool const hasErrored = _error->PendingError();
    _error->MergeWithStack();
    if (hasErrored)
@@ -364,7 +431,7 @@ bool OpenMaybeClearSignedFile(std::string const &ClearSignedFileName, FileFd &Me
       free(message);
       return _error->Errno("mkstemp", "Couldn't create temporary file to work with %s", ClearSignedFileName.c_str());
    }
-   // we have the fd, thats enough for us
+   // we have the fd, that's enough for us
    unlink(message);
    free(message);
 
