@@ -11,25 +11,45 @@
 // Include Files							/*{{{*/
 #include <config.h>
 
+#include <apt-pkg/aptconfiguration.h>
 #include <apt-pkg/deblistparser.h>
 #include <apt-pkg/debsrcrecords.h>
 #include <apt-pkg/error.h>
-#include <apt-pkg/strutl.h>
-#include <apt-pkg/aptconfiguration.h>
+#include <apt-pkg/gpgv.h>
+#include <apt-pkg/hashes.h>
 #include <apt-pkg/srcrecords.h>
+#include <apt-pkg/strutl.h>
+#include <apt-pkg/tagfile-keys.h>
 #include <apt-pkg/tagfile.h>
 
-#include <ctype.h>
-#include <stdlib.h>
-#include <string.h>
 #include <algorithm>
 #include <string>
 #include <vector>
+#include <ctype.h>
+#include <stdlib.h>
+#include <string.h>
 									/*}}}*/
 
 using std::max;
 using std::string;
 
+debSrcRecordParser::debSrcRecordParser(std::string const &File,pkgIndexFile const *Index)
+   : Parser(Index), d(NULL), Tags(&Fd), iOffset(0), Buffer(NULL)
+{
+   if (File.empty() == false)
+   {
+      if (Fd.Open(File, FileFd::ReadOnly, FileFd::Extension))
+	 Tags.Init(&Fd, 102400);
+   }
+}
+std::string debSrcRecordParser::Package() const				/*{{{*/
+{
+   auto const name = Sect.Find(pkgTagSection::Key::Package);
+   if (iIndex != nullptr || name.empty() == false)
+      return name.to_string();
+   return Sect.Find(pkgTagSection::Key::Source).to_string();
+}
+									/*}}}*/
 // SrcRecordParser::Binaries - Return the binaries field		/*{{{*/
 // ---------------------------------------------------------------------
 /* This member parses the binaries field into a pair of class arrays and
@@ -40,9 +60,9 @@ using std::string;
 const char **debSrcRecordParser::Binaries()
 {
    const char *Start, *End;
-   if (Sect.Find("Binary", Start, End) == false)
+   if (Sect.Find(pkgTagSection::Key::Binary, Start, End) == false)
       return NULL;
-   for (; isspace(*Start) != 0; ++Start);
+   for (; isspace_ascii(*Start) != 0; ++Start);
    if (Start >= End)
       return NULL;
 
@@ -53,14 +73,18 @@ const char **debSrcRecordParser::Binaries()
    char* bin = Buffer;
    do {
       char* binStartNext = strchrnul(bin, ',');
-      char* binEnd = binStartNext - 1;
-      for (; isspace(*binEnd) != 0; --binEnd)
-	 binEnd = '\0';
+      // Found a comma, clean up any space before it
+      if (binStartNext > Buffer) {
+	 char* binEnd = binStartNext - 1;
+	 for (; binEnd > Buffer && isspace_ascii(*binEnd) != 0; --binEnd)
+	    *binEnd = 0;
+      }
       StaticBinList.push_back(bin);
       if (*binStartNext != ',')
 	 break;
       *binStartNext = '\0';
-      for (bin = binStartNext + 1; isspace(*bin) != 0; ++bin);
+      for (bin = binStartNext + 1; isspace_ascii(*bin) != 0; ++bin)
+         ;
    } while (*bin != '\0');
    StaticBinList.push_back(NULL);
 
@@ -76,41 +100,58 @@ const char **debSrcRecordParser::Binaries()
 bool debSrcRecordParser::BuildDepends(std::vector<pkgSrcRecords::Parser::BuildDepRec> &BuildDeps,
 					bool const &ArchOnly, bool const &StripMultiArch)
 {
-   unsigned int I;
-   const char *Start, *Stop;
-   BuildDepRec rec;
-   const char *fields[] = {"Build-Depends", 
-                           "Build-Depends-Indep",
-			   "Build-Conflicts",
-			   "Build-Conflicts-Indep"};
-
    BuildDeps.clear();
 
-   for (I = 0; I < 4; I++) 
+   pkgTagSection::Key const fields[] = {
+      pkgTagSection::Key::Build_Depends,
+      pkgTagSection::Key::Build_Depends_Indep,
+      pkgTagSection::Key::Build_Conflicts,
+      pkgTagSection::Key::Build_Conflicts_Indep,
+      pkgTagSection::Key::Build_Depends_Arch,
+      pkgTagSection::Key::Build_Conflicts_Arch,
+   };
+   for (unsigned short I = 0; I < sizeof(fields) / sizeof(fields[0]); ++I)
    {
-      if (ArchOnly && (I == 1 || I == 3))
-         continue;
+      if (ArchOnly && (fields[I] == pkgTagSection::Key::Build_Depends_Indep || fields[I] == pkgTagSection::Key::Build_Conflicts_Indep))
+	 continue;
 
+      const char *Start, *Stop;
       if (Sect.Find(fields[I], Start, Stop) == false)
          continue;
-      
+
+      if (Start == Stop)
+	 continue;
+
       while (1)
       {
-         Start = debListParser::ParseDepends(Start, Stop, 
-		     rec.Package,rec.Version,rec.Op,true,StripMultiArch,true);
-	 
-         if (Start == 0) 
-            return _error->Error("Problem parsing dependency: %s", fields[I]);
+	 BuildDepRec rec;
+	 Start = debListParser::ParseDepends(Start, Stop,
+					     rec.Package, rec.Version, rec.Op, true, StripMultiArch, true);
+
+	 if (Start == 0)
+	    return _error->Error("Problem parsing dependency: %s", BuildDepType(I));
 	 rec.Type = I;
 
-	 if (rec.Package != "")
-   	    BuildDeps.push_back(rec);
-	 
-   	 if (Start == Stop) 
+	 // We parsed a package that was ignored (wrong architecture restriction
+	 // or something).
+	 if (rec.Package.empty())
+	 {
+	    // If we are in an OR group, we need to set the "Or" flag of the
+	    // previous entry to our value.
+	    if (BuildDeps.empty() == false && (BuildDeps[BuildDeps.size() - 1].Op & pkgCache::Dep::Or) == pkgCache::Dep::Or)
+	    {
+	       BuildDeps[BuildDeps.size() - 1].Op &= ~pkgCache::Dep::Or;
+	       BuildDeps[BuildDeps.size() - 1].Op |= (rec.Op & pkgCache::Dep::Or);
+	    }
+	 } else {
+	    BuildDeps.emplace_back(std::move(rec));
+	 }
+
+	 if (Start == Stop)
 	    break;
-      }	 
+      }
    }
-   
+
    return true;
 }
 									/*}}}*/
@@ -146,7 +187,7 @@ bool debSrcRecordParser::Files2(std::vector<pkgSrcRecords::File2> &List)
    List.clear();
 
    // Stash the / terminated directory prefix
-   string Base = Sect.FindS("Directory");
+   std::string Base = Sect.Find(pkgTagSection::Key::Directory).to_string();
    if (Base.empty() == false && Base[Base.length()-1] != '/')
       Base += '/';
 
@@ -177,6 +218,15 @@ bool debSrcRecordParser::Files2(std::vector<pkgSrcRecords::File2> &List)
 	       ParseQuoteWord(C, path) == false)
 	    return _error->Error("Error parsing file record in %s of source package %s", checksumField.c_str(), Package().c_str());
 
+	 if (iIndex == nullptr && checksumField == "Files")
+	 {
+	    // the Files field has a different format than the rest in deb-changes files
+	    std::string ignore;
+	    if (ParseQuoteWord(C, ignore) == false ||
+		  ParseQuoteWord(C, path) == false)
+	       return _error->Error("Error parsing file record in %s of source package %s", checksumField.c_str(), Package().c_str());
+	 }
+
 	 HashString const hashString(*type, hash);
 	 if (Base.empty() == false)
 	    path = Base + path;
@@ -190,16 +240,12 @@ bool debSrcRecordParser::Files2(std::vector<pkgSrcRecords::File2> &List)
 	 // we have it already, store the new hash and be done
 	 if (file != List.end())
 	 {
-#if __GNUC__ >= 4
-	// set for compatibility only, so warn users not us
-	#pragma GCC diagnostic push
-	#pragma GCC diagnostic ignored "-Wdeprecated-declarations"
-#endif
 	    if (checksumField == "Files")
+	    {
+	       APT_IGNORE_DEPRECATED_PUSH
 	       file->MD5Hash = hash;
-#if __GNUC__ >= 4
-	#pragma GCC diagnostic pop
-#endif
+	       APT_IGNORE_DEPRECATED_POP
+	    }
 	    // an error here indicates that we have two different hashes for the same file
 	    if (file->Hashes.push_back(hashString) == false)
 	       return _error->Error("Error parsing checksum in %s of source package %s", checksumField.c_str(), Package().c_str());
@@ -211,18 +257,13 @@ bool debSrcRecordParser::Files2(std::vector<pkgSrcRecords::File2> &List)
 	 F.Path = path;
 	 F.FileSize = strtoull(size.c_str(), NULL, 10);
 	 F.Hashes.push_back(hashString);
+	 F.Hashes.FileSize(F.FileSize);
 
-#if __GNUC__ >= 4
-	// set for compatibility only, so warn users not us
-	#pragma GCC diagnostic push
-	#pragma GCC diagnostic ignored "-Wdeprecated-declarations"
-#endif
+	 APT_IGNORE_DEPRECATED_PUSH
 	 F.Size = F.FileSize;
 	 if (checksumField == "Files")
 	    F.MD5Hash = hash;
-#if __GNUC__ >= 4
-	#pragma GCC diagnostic pop
-#endif
+	 APT_IGNORE_DEPRECATED_POP
 
 	 // Try to guess what sort of file it is we are getting.
 	 string::size_type Pos = F.Path.length()-1;
@@ -264,3 +305,21 @@ debSrcRecordParser::~debSrcRecordParser()
    free(Buffer);
 }
 									/*}}}*/
+
+
+debDscRecordParser::debDscRecordParser(std::string const &DscFile, pkgIndexFile const *Index)
+   : debSrcRecordParser("", Index)
+{
+   // support clear signed files
+   if (OpenMaybeClearSignedFile(DscFile, Fd) == false)
+   {
+      _error->Error("Failed to open %s", DscFile.c_str());
+      return;
+   }
+
+   // re-init to ensure the updated Fd is used
+   Tags.Init(&Fd, pkgTagFile::SUPPORT_COMMENTS);
+   // read the first (and only) record
+   Step();
+
+}
