@@ -8,34 +8,33 @@
    ##################################################################### */
 									/*}}}*/
 // Include Files							/*{{{*/
-#include<config.h>
+#include <config.h>
 
+#include <apt-pkg/aptconfiguration.h>
+#include <apt-pkg/cachefile.h>
+#include <apt-pkg/cacheset.h>
+#include <apt-pkg/configuration.h>
 #include <apt-pkg/depcache.h>
-#include <apt-pkg/versionmatch.h>
-#include <apt-pkg/version.h>
 #include <apt-pkg/error.h>
 #include <apt-pkg/fileutl.h>
-#include <apt-pkg/strutl.h>
-#include <apt-pkg/configuration.h>
-#include <apt-pkg/aptconfiguration.h>
-#include <apt-pkg/tagfile.h>
-#include <apt-pkg/progress.h>
-#include <apt-pkg/cacheset.h>
-#include <apt-pkg/pkgcache.h>
-#include <apt-pkg/cacheiterators.h>
-#include <apt-pkg/prettyprinters.h>
-#include <apt-pkg/cachefile.h>
 #include <apt-pkg/macros.h>
+#include <apt-pkg/pkgcache.h>
+#include <apt-pkg/prettyprinters.h>
+#include <apt-pkg/progress.h>
+#include <apt-pkg/strutl.h>
+#include <apt-pkg/tagfile.h>
+#include <apt-pkg/version.h>
+#include <apt-pkg/versionmatch.h>
 
-#include <stdio.h>
-#include <string.h>
+#include <algorithm>
+#include <iostream>
 #include <list>
+#include <set>
 #include <string>
 #include <utility>
 #include <vector>
-#include <algorithm>
-#include <iostream>
-#include <set>
+#include <stdio.h>
+#include <string.h>
 
 #include <sys/stat.h>
 
@@ -183,7 +182,7 @@ bool pkgDepCache::readStateFile(OpProgress * const Prog)		/*{{{*/
    FileFd state_file;
    string const state = _config->FindFile("Dir::State::extended_states");
    if(RealFileExists(state)) {
-      state_file.Open(state, FileFd::ReadOnly);
+      state_file.Open(state, FileFd::ReadOnly, FileFd::Extension);
       off_t const file_size = state_file.Size();
       if(Prog != NULL)
 	 Prog->OverallProgress(0, file_size, 1,
@@ -245,16 +244,16 @@ bool pkgDepCache::writeStateFile(OpProgress * const /*prog*/, bool const Install
    // if it does not exist, create a empty one
    if(!RealFileExists(state))
    {
-      StateFile.Open(state, FileFd::WriteAtomic);
+      StateFile.Open(state, FileFd::WriteAtomic, FileFd::Extension);
       StateFile.Close();
    }
 
    // open it
-   if(!StateFile.Open(state, FileFd::ReadOnly))
+   if (!StateFile.Open(state, FileFd::ReadOnly, FileFd::Extension))
       return _error->Error(_("Failed to open StateFile %s"),
 			   state.c_str());
 
-   FileFd OutFile(state, FileFd::ReadWrite | FileFd::Atomic);
+   FileFd OutFile(state, FileFd::ReadWrite | FileFd::Atomic, FileFd::Extension);
    if (OutFile.IsOpen() == false || OutFile.Failed() == true)
       return _error->Error(_("Failed to write temporary StateFile %s"), state.c_str());
 
@@ -274,23 +273,27 @@ bool pkgDepCache::writeStateFile(OpProgress * const /*prog*/, bool const Install
 	    continue;
 	 StateCache const &P = PkgState[pkg->ID];
 	 bool newAuto = (P.Flags & Flag::Auto);
-	 // skip not installed or now-removed ones if requested
+	 // reset to default (=manual) not installed or now-removed ones if requested
 	 if (InstalledOnly && (
 	     (pkg->CurrentVer == 0 && P.Mode != ModeInstall) ||
 	     (pkg->CurrentVer != 0 && P.Mode == ModeDelete)))
+	    newAuto = false;
+	 if (newAuto == false)
 	 {
 	    // The section is obsolete if it contains no other tag
-	    unsigned int const count = section.Count();
+	    auto const count = section.Count();
 	    if (count < 2 ||
 		(count == 2 && section.Exists("Auto-Installed")) ||
 		(count == 3 && section.Exists("Auto-Installed") && section.Exists("Architecture")))
+	    {
+	       if(debug_autoremove)
+		  std::clog << "Drop obsolete section with " << count << " fields for " << APT::PrettyPkg(this, pkg) << std::endl;
 	       continue;
-	    else
-	       newAuto = false;
+	    }
 	 }
-	 if(_config->FindB("Debug::pkgAutoRemove",false))
-	    std::clog << "Update existing AutoInstall info: " 
-		      << pkg.FullName() << std::endl;
+
+	 if(debug_autoremove)
+	    std::clog << "Update existing AutoInstall to " << newAuto << " for " << APT::PrettyPkg(this, pkg) << std::endl;
 
 	 std::vector<pkgTagSection::Tag> rewrite;
 	 rewrite.push_back(pkgTagSection::Tag::Rewrite("Architecture", pkg.Arch()));
@@ -307,7 +310,7 @@ bool pkgDepCache::writeStateFile(OpProgress * const /*prog*/, bool const Install
       if(P.Flags & Flag::Auto) {
 	 if (pkgs_seen.find(pkg.FullName()) != pkgs_seen.end()) {
 	    if(debug_autoremove)
-	       std::clog << "Skipping already written " << pkg.FullName() << std::endl;
+	       std::clog << "Skipping already written " << APT::PrettyPkg(this, pkg) << std::endl;
 	    continue;
 	 }
 	 // skip not installed ones if requested
@@ -315,14 +318,11 @@ bool pkgDepCache::writeStateFile(OpProgress * const /*prog*/, bool const Install
 	     (pkg->CurrentVer == 0 && P.Mode != ModeInstall) ||
 	     (pkg->CurrentVer != 0 && P.Mode == ModeDelete)))
 	    continue;
-	 const char* const pkgarch = pkg.Arch();
-	 if (strcmp(pkgarch, "all") == 0)
-	    continue;
 	 if(debug_autoremove)
-	    std::clog << "Writing new AutoInstall: " << pkg.FullName() << std::endl;
+	    std::clog << "Writing new AutoInstall: " << APT::PrettyPkg(this, pkg) << std::endl;
 	 std::string stanza = "Package: ";
 	 stanza.append(pkg.Name())
-	      .append("\nArchitecture: ").append(pkgarch)
+	      .append("\nArchitecture: ").append(pkg.Arch())
 	      .append("\nAuto-Installed: 1\n\n");
 	 if (OutFile.Write(stanza.c_str(), stanza.length()) == false)
 	    return false;
@@ -786,7 +786,7 @@ bool pkgDepCache::MarkKeep(PkgIterator const &Pkg, bool Soft, bool FromUser,
 
    ActionGroup group(*this);
 
-#if 0 // reseting the autoflag here means we lose the 
+#if 0 // resetting the autoflag here means we lose the 
       // auto-mark information if a user selects a package for removal
       // but changes  his mind then and sets it for keep again
       // - this makes sense as default when all Garbage dependencies
@@ -1078,7 +1078,7 @@ bool pkgDepCache::MarkInstall(PkgIterator const &Pkg,bool AutoInst,
 
    StateCache &P = PkgState[Pkg->ID];
 
-   // See if there is even any possible instalation candidate
+   // See if there is even any possible installation candidate
    if (P.CandidateVer == 0)
       return false;
 
@@ -1559,6 +1559,13 @@ bool pkgDepCache::SetCandidateRelease(pkgCache::VerIterator TargetVer,
 	 // virtual packages can't be a solution
 	 if (P.end() == true || (P->ProvidesList == 0 && P->VersionList == 0))
 	    continue;
+	 // if its already installed, check if this one is good enough
+	 pkgCache::VerIterator const Now = P.CurrentVer();
+	 if (Now.end() == false && Start.IsSatisfied(Now))
+	 {
+	    itsFine = true;
+	    break;
+	 }
 	 pkgCache::VerIterator const Cand = PkgState[P->ID].CandidateVerIter(*this);
 	 // no versioned dependency - but is it installable?
 	 if (Start.TargetVer() == 0 || Start.TargetVer()[0] == '\0')
@@ -1780,9 +1787,9 @@ bool pkgDepCache::Policy::IsImportantDep(DepIterator const &Dep) const
    {
       if (InstallRecommends)
 	 return true;
-      // we suport a special mode to only install-recommends for certain
+      // we support a special mode to only install-recommends for certain
       // sections
-      // FIXME: this is a meant as a temporarly solution until the
+      // FIXME: this is a meant as a temporary solution until the
       //        recommends are cleaned up
       const char *sec = Dep.ParentVer().Section();
       if (sec && ConfigValueInSubTree("APT::Install-Recommends-Sections", sec))
@@ -1795,11 +1802,11 @@ bool pkgDepCache::Policy::IsImportantDep(DepIterator const &Dep) const
 }
 									/*}}}*/
 // Policy::GetPriority - Get the priority of the package pin		/*{{{*/
-APT_CONST signed short pkgDepCache::Policy::GetPriority(pkgCache::PkgIterator const &/*Pkg*/)
+APT_PURE signed short pkgDepCache::Policy::GetPriority(pkgCache::PkgIterator const &/*Pkg*/)
 { return 0; }
-APT_CONST signed short pkgDepCache::Policy::GetPriority(pkgCache::VerIterator const &/*Ver*/, bool /*ConsiderFiles*/)
+APT_PURE signed short pkgDepCache::Policy::GetPriority(pkgCache::VerIterator const &/*Ver*/, bool /*ConsiderFiles*/)
 { return 0; }
-APT_CONST signed short pkgDepCache::Policy::GetPriority(pkgCache::PkgFileIterator const &/*File*/)
+APT_PURE signed short pkgDepCache::Policy::GetPriority(pkgCache::PkgFileIterator const &/*File*/)
 { return 0; }
 									/*}}}*/
 pkgDepCache::InRootSetFunc *pkgDepCache::GetRootSetFunc()		/*{{{*/
